import numpy as np
from pypet import Trajectory
from pypet.brian2 import Brian2MonitorResult
from tqdm import tqdm

from scripts.spatial_network.head_direction_index_over_noise_scale import calculate_rates
from scripts.spatial_network.run_entropy_maximisation_orientation_map import DATA_FOLDER, TRAJ_NAME


def get_spike_train_dictionary(number_of_neurons, spike_times, neuron_indices):
    spike_train_dict = {}
    for neuron_idx in range(number_of_neurons):
        spike_train_dict[neuron_idx] = []

    for neuron_idx, t in zip(neuron_indices, spike_times):
        spike_train_dict[neuron_idx].append(t)
    print(spike_train_dict)
    return spike_train_dict


def get_firing_rate_per_cell_and_direction(traj, run_name):
    traj.f_set_crun(run_name)
    firing_rate_dict = {}
    direction_names = ["dir{:d}".format(idx) for idx in range(traj.input.number_of_directions)]
    for idx in range(traj.N_E):
        firing_rate_dict[idx] = []
    for direction in direction_names:
        ex_spike_trains = get_spike_train_dictionary(traj.N_E, traj.results.runs[run_name][direction].spikes.e.t,
                                                     traj.results.runs[run_name][
                                                         direction].spikes.e.i)
        ex_spike_rates = calculate_rates(ex_spike_trains.values())
        for idx, spike_rate in enumerate(ex_spike_rates):
            firing_rate_dict[idx].append(spike_rate)
    traj.f_restore_default()
    return firing_rate_dict


def get_head_direction_indices(directions, firing_rate_dict):
    n_exc_neurons = len(firing_rate_dict)
    n_directions = len(directions)

    tuning_vectors = np.ndarray((n_exc_neurons,n_directions,2))

    for ex_id, ex_rates in firing_rate_dict.items():
        rate_sum = 0.
        for dir_id, dir in enumerate(directions):
            tuning_vectors[ex_id, dir_id] = np.array([np.cos(dir), np.sin(dir)]) * ex_rates[dir_id]
            rate_sum += ex_rates[dir_id]
        tuning_vectors[ex_id, :] /= rate_sum

    tuning_vectors_summed = np.sum(tuning_vectors, axis=1)

    head_direction_indices = np.array([np.linalg.norm(v) for v in tuning_vectors_summed])

    return head_direction_indices


def get_runs_with_circular_morphology(traj):
    filtered_indices = traj.f_find_idx(('parameters.long_axis', 'parameters.short_axis'), lambda r1, r2: r1 == r2)
    return filtered_indices


if __name__ == "__main__":
    traj = Trajectory(TRAJ_NAME, add_time=False, dynamic_imports=Brian2MonitorResult)
    NO_LOADING = 0
    FULL_LOAD = 2
    # TODO: Why no loading of results? Because of dynamic loading?
    traj.f_load(filename=DATA_FOLDER + TRAJ_NAME + ".hdf5", load_parameters=FULL_LOAD, load_results=NO_LOADING)

    traj.v_auto_load = True

    correlation_lengths = traj.f_get('correlation_length').f_get_range()
    long_axis = traj.f_get('long_axis').f_get_range()
    short_axis = traj.f_get('short_axis').f_get_range()
    # TODO: Again, maybe directions as parameters
    directions = np.linspace(-np.pi, np.pi, traj.input.number_of_directions)

    circular_indices = list(get_runs_with_circular_morphology(traj))

    for idx, run_name in enumerate(traj.f_get_run_names()):
        firing_rate_dict = get_firing_rate_per_cell_and_direction(traj, run_name)
        print(firing_rate_dict)
        head_direction_indices = get_head_direction_indices(directions, firing_rate_dict)
        traj.f_set_crun(run_name)
        print("Circle" if idx in circular_indices else "Ellipsoid" )
        print("Corr length {:.1f}".format(traj.orientation_map.correlation_length))
        print("Long axis {:.1f}".format(traj.long_axis))
        print("Mean HDI {:.1f}".format(np.mean(head_direction_indices)))
        print()

    traj.f_restore_default()
